/* BASIC/LL runtime-library */

#include "libbasicll.h"

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <io.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>

#ifndef SEEK_CUR
#define SEEK_CUR 1
#endif

/* PUBLIC */

int basic_argc;
char **basic_args;
int basic_exit_code;
intptr_t basic_current_line;
jmp_buf basic_on_exit;

/* CONTROL */

static INTEGER error_number = 0;
static INTEGER error_line = 0;

#ifdef BASIC_CONTROL_STACK

struct basic_control_stack_t {
	struct basic_control_stack_t *next;
	enum basic_control_kind_t kind;
	void *control;
	jmp_buf start;
};

static struct basic_control_stack_t *basic_control_stack = NULL;

jmp_buf *enter_control(enum basic_control_kind_t kind, void *control)
{
	struct basic_control_stack_t *item;
	item = malloc(sizeof(struct basic_control_stack_t));
	item->next = basic_control_stack;
	item->kind = kind;
	item->control = control;
	basic_control_stack = item;
	/* caller should call setjmp after enter_control */
	return &item->start;
}

void leave_control(enum basic_control_kind_t kind, void *control)
{
	struct basic_control_stack_t *next;
	INTEGER exit;
	for(;;){
		if(basic_control_stack == NULL){
			ERROR(next_without_for);
		}
		next = basic_control_stack->next;
		exit = basic_control_stack->kind == kind && basic_control_stack->control == control;
		free(basic_control_stack);
		basic_control_stack = next;
		if(exit){
			break;
		}
	}
}

void back_control(enum basic_control_kind_t kind, void *control)
{
	struct basic_control_stack_t *next;
	for(;;){
		if(basic_control_stack == NULL){
			ERROR(next_without_for);
		}
		if(basic_control_stack->kind == kind && basic_control_stack->control == control){
			break;
		}
		next = basic_control_stack->next;
		free(basic_control_stack);
		basic_control_stack = next;
	}
	longjmp(basic_control_stack->start, 1);
}

#endif

void END(void)
{
	error_number = 0;
	longjmp(basic_on_exit, 1);
}

INTEGER ERR(void)
{
	return error_number;
}

INTEGER ERL(void)
{
	return error_line;
}

void ERROR(INTEGER number)
{
	if(number == 0){
		number = illegal_function_call;
	}
	error_number = number;
	error_line = basic_current_line;
	longjmp(basic_on_exit, number);
}

/* MEMORY */

void *allocate(INTEGER size)
{
	return malloc(size);
}

void *reallocate(void *memory, INTEGER size)
{
	return realloc(memory, size);
}

void deallocate(void *memory)
{
	free(memory);
}

/* STRING */

void string_addref(STRING s)
{
	if(s != NULL){
		intptr_t *ref_count = (intptr_t *)(s - sizeof(intptr_t) * 2);
		if(*ref_count != -1){
			++ *ref_count;
		}
	}
}

void string_release(STRING s)
{
	if(s != NULL){
		intptr_t *ref_count = (intptr_t *)(s - sizeof(intptr_t) * 2);
		if(*ref_count != -1){
			-- *ref_count;
			if(*ref_count == 0){
				free(ref_count);
			}
		}
	}
}

unsigned char *new_string(unsigned char const *data, intptr_t length)
{
	unsigned char *result;
	intptr_t *new_ref_count;
	intptr_t *new_length;
	if(length == 0){
		result = NULL;
	}else{
		new_ref_count = malloc(length + (sizeof(intptr_t) * 2 + 1));
		if(new_ref_count == NULL){
			ERROR(out_of_memory);
		}
		*new_ref_count = 1;
		new_length = new_ref_count + 1;
		*new_length = length;
		result = (unsigned char *)(new_length + 1);
		result[length] = 0;
		if(data != NULL){
			memcpy(result, data, length);
		}
	}
	return result;
}

void unique_string(STRING *s)
{
	STRING old;
	if(s != NULL){
		intptr_t *ref_count = (intptr_t *)(*s - sizeof(intptr_t) * 2);
		if(*ref_count >= 2 || *ref_count == -1){
			intptr_t *length = (intptr_t *)(*s - sizeof(intptr_t));
			old = *s;
			*s = new_string(old, *length);
			string_release(old);
		}
	}
}

void set_string(STRING *s, unsigned char const *data, intptr_t length)
{
	STRING new = new_string(data, length);
	string_release(*s);
	*s = new;
}

void set_length(STRING *s, intptr_t length)
{
	INTEGER s_length = LEN(*s);
	unsigned char *new;
	intptr_t *ref_count;
	intptr_t *new_length;
	if(s_length == length){
		unique_string(s);
	}else{
		ref_count = (intptr_t *)(*s - sizeof(intptr_t) * 2);
		if(*s != NULL && *(ref_count = (intptr_t *)(*s - sizeof(intptr_t) * 2)) == 1){
			intptr_t *new_ref_count = realloc(ref_count, length + (sizeof(intptr_t) * 2 + 1));
			if(new_ref_count == NULL){
				ERROR(out_of_memory);
			}
			new_length = new_ref_count + 1;
			*new_length = length;
			new = (unsigned char *)(new_length + 1);
			new[length] = 0;
			*s = new;
		}else{
			new = new_string(NULL, length);
			memcpy((char *)new, *s, s_length);
			string_release(*s);
			*s = new;
		}
	}
}

INTEGER LEN(STRING s)
{
	INTEGER result;
	if(s != NULL){
		intptr_t *length = (intptr_t *)(s - sizeof(intptr_t));
		result = *length;
	}else{
		result = 0;
	}
	return result;
}

INTEGER string_compare(STRING left, STRING right)
{
	INTEGER result;
	INTEGER l_length = LEN(left);
	INTEGER r_length = LEN(right);
	result = memcmp(left, right, (l_length < r_length) ? l_length : r_length);
	if(result == 0){
		result = l_length - r_length;
	}
	return result;
}

STRING string_concatenate(STRING left, STRING right)
{
	STRING result;
	INTEGER l_length;
	INTEGER r_length;
	INTEGER new_length;
	unsigned char *new;
	if((r_length = LEN(right)) == 0){
		result = left;
		string_addref(result);
	}else if((l_length = LEN(left)) == 0){
		result = right;
		string_addref(result);
	}else{
		new_length = l_length + r_length;
		if(new_length <= 0){
			ERROR(string_too_long); /* overflow */
		}
		new = new_string(NULL, new_length);
		memcpy(new, left, l_length);
		memcpy(new + l_length, right, r_length);
		result = new;
	}
	return result;
}

CHARACTER const *STRPTR(STRING s)
{
	static unsigned char null_string[1] = {0};
	
	unsigned char const *result;
	if(s != NULL){
		unique_string(&s);
		result = s;
	}else{
		result = null_string;
	}
	return result;
}

STRING MID(STRING s, INTEGER index, INTEGER length)
{
	STRING result;
	unsigned char const *s_end = s + LEN(s);
	unsigned char const *start;
	unsigned char const *end;
	start = s + index - 1;
	if(length >= 0){
		end = start + length;
	}else{
		end = start;
		start = end + length;
	}
	if(start > s_end){
		start = s_end;
	}else if(start < s){
		start = s;
	}
	if(end > s_end){
		end = s_end;
	}else if(end < s){
		end = s;
	}
	if(start == s && end == s_end){
		result = s;
		string_addref(result);
	}else{
		result = new_string(start, end - start);
	}
	return result;
}

STRING LEFT(STRING s, INTEGER length)
{
	return MID(s, 1, length);
}

STRING RIGHT(STRING s, INTEGER length)
{
	return MID(s, 1 + LEN(s), - length);
}

STRING SPACE(INTEGER count)
{
	unsigned char *result = NULL;
	if(count > 0){
		new_string(NULL, count);
		memset(result, ' ', count);
	}
	return result;
}

STRING fnSTRING(STRING s, INTEGER count)
{
	INTEGER length = LEN(s);
	INTEGER i;
	unsigned char *result = new_string(NULL, length * count);
	for(i = 0; i < count; ++i){
		memcpy(result + i * length, s, length);
	}
	return result;
}

STRING UCASE(STRING s)
{
	INTEGER length = LEN(s);
	INTEGER i;
	STRING result = s;
	string_addref(result);
	for(i = 0; i < length; ++i){
		if(result[i] >= 'a' && result[i] <= 'z'){
			unique_string(&result);
			((char *)result)[i] -= 'a' - 'A';
		}
	}
	return result;
}

STRING LCASE(STRING s)
{
	INTEGER length = LEN(s);
	INTEGER i;
	STRING result = s;
	string_addref(result);
	for(i = 0; i < length; ++i){
		if(result[i] >= 'A' && result[i] <= 'Z'){
			unique_string(&result);
			unique_string(&result);
			((char *)result)[i] += 'a' - 'A';
		}
	}
	return result;
}

void using_string(STRING *output, STRING format, INTEGER *index, STRING value)
{
	STRING s = NULL;
	INTEGER column = 0;
	INTEGER start = *index;
	INTEGER length = LEN(format);
	for(;;){
		if(*index >= length){
			ERROR(missing_operand);
		}
		if(format[*index] == '#'){
			break;
		}
		if(format[*index] == '@'){
			column = -1;
			break;
		}
		++ *index;
	}
	s = MID(format, start + 1, *index - start);
	string_append(output, s);
	if(column < 0){
		++ *index;
		column = LEN(value);
	}else{
		do{
			++ column;
			++ *index;
		}while(format[*index] == '#');
	}
	string_append(output, value);
	string_release(s);
	s = SPACE(column - LEN(value));
	string_append(output, s);
	string_release(s);
}

void using_double(STRING *output, STRING format, INTEGER *index, DOUBLE value)
{
	unsigned char buffer[1024];
	STRING s = NULL;
	int zero_column = 0;
	int sgn = 0;
	int sgn_pos = 0;
	int has_point = 0;
	int prec = 0;
	int integer_column = 0;
	int need_integer_column;
	int minus;
	int i;
	DOUBLE x;
	DOUBLE integer_part;
	int mod;
	INTEGER start = *index;
	INTEGER length = LEN(format);
	CHARACTER c;
	for(;;){
		if(*index >= length){
			ERROR(missing_operand);
		}
		c = format[*index];
		if(c == '#' || ((c == '+' || c == '-') && (format[*index + 1] == '#' || format[*index + 1] == '0' || format[*index + 1] == '.'))){
			break;
		}
		++ *index;
	}
	s = MID(format, start + 1, *index - start);
	string_append(output, s);
	while(format[*index] == '#' || format[*index] == '+' || format[*index] == '-'){
		if(format[*index] == '+'){
			if(sgn != 0){
				break;
			}
			sgn = +1;
			sgn_pos = integer_column;
		}
		if(format[*index] == '-'){
			if(sgn != 0){
				break;
			}
			sgn = -1;
			sgn_pos = integer_column;
		}
		++ integer_column;
		++ *index;
	}
	while(format[*index] == '0'){
		++ integer_column;
		++ zero_column;
		++ *index;
	}
	if(format[*index] == '.'){
		has_point = 1;
		++ *index;
		while(format[*index] == '#'){
			++ prec;
			++ *index;
		}
	}
	minus = 0;
	if(value < 0){
		minus = 1;
		value = - value;
	}
	if(! has_point){
#ifdef __BORLANDC__
		value = floor(value + 0.5); /* Borland C doesn't have nearbyint. */
#else
		value = nearbyint(value); /* banker's rounding */
#endif
	}
	need_integer_column = (value < 1.0 && ! has_point);
	x = 1.0F;
	while(x <= value){
		++ need_integer_column;
		x = x * 10.0F;
	}
	integer_part = floor(value);
	value = value - integer_part;
	if(integer_column < need_integer_column + (minus || sgn > 0)){
		integer_column = need_integer_column + (minus || sgn > 0);
	}
	for(i = integer_column - 1; i >= integer_column - need_integer_column; -- i){
		mod = fmod(integer_part, 10.0F);
		buffer[i] = '0' + mod;
		integer_part = floor(integer_part / 10.0F);
	}
	while(i >= integer_column - zero_column){
		buffer[i] = '0';
		-- i;
	}
	while(sgn != 0 && sgn_pos < i){
		buffer[i] = ' ';
		--i;
	}
	if(minus || sgn > 0){
		buffer[i] = minus ? '-' : '+';
		-- i;
	}
	while(i >= 0){
		buffer[i] = ' ';
		-- i;
	}
	if(has_point){
		buffer[integer_column] = '.';
		for(i = 1; i <= prec; ++i){
			value = value * 10.0F;
			mod = (int)value;
			value = value - floor(value);
			if(i == prec){
				if(value >= 0.5){
					/* crazy implementation for rounding */
					if(mod < 9){
						++ mod;
					}
				}
			}
			buffer[integer_column + i] = '0' + (int)mod;
		}
	}
	set_string(&s, buffer, integer_column + has_point + prec);
	string_append(output, s);
	string_release(s);
}

void using_rest(STRING *output, STRING format, INTEGER *index)
{
	INTEGER length = LEN(format);
	STRING s = MID(format, *index + 1, length);
	string_append(output, s);
	*index = length;
	string_release(s);
}

/* DATA CONVERTION */

STRING MKI(INTEGER value)
{
	return new_string((unsigned char *)&value, sizeof(INTEGER));
}

STRING MKS(SINGLE value)
{
	return new_string((unsigned char *)&value, sizeof(SINGLE));
}

STRING MKD(DOUBLE value)
{
	return new_string((unsigned char *)&value, sizeof(DOUBLE));
}

INTEGER CVI(STRING s)
{
	if(LEN(s) < (INTEGER)sizeof(INTEGER)){
		ERROR(illegal_function_call);
	}
	return *(INTEGER *)s;
}

SINGLE CVS(STRING s)
{
	if(LEN(s) < (INTEGER)sizeof(SINGLE)){
		ERROR(illegal_function_call);
	}
	return *(SINGLE *)s;
}

DOUBLE CVD(STRING s)
{
	if(LEN(s) < (INTEGER)sizeof(DOUBLE)){
		ERROR(illegal_function_call);
	}
	return *(DOUBLE *)s;
}

STRING STR(DOUBLE x)
{
	unsigned char *result;
	unsigned char buffer[1024];
	int ip = 0;
	int pp = 0;
	int length;
	int minus = 0;
	int prec = 0;
	double integer_part;
	int mod;
	if(x < 0){
		minus = 1;
		x = - x;
	}
	integer_part = floor(x);
	x = x - integer_part;
	while(integer_part != 0.0F){
		mod = fmod(integer_part, 10.0F);
		buffer[ip] = '0' + mod;
		integer_part = floor(integer_part / 10.0F);
		++ ip;
		if(mod != 0 || prec != 0){
			++ prec;
		}
	}
	if(x != 0.0){
		pp = ip;
		buffer[pp] = '.';
		++ pp;
		do{
			x = x * 10.0F;
			mod = (int)x;
			x = x - floor(x);
			if(mod != 0 || prec != 0){
				++ prec;
				if(prec >= 16){ /* limit of 64bit double real number */
					if(x >= 0.5){
						/* crazy implementation for rounding */
						if(mod < 9){
							++ mod;
						}
					}
					x = 0.0F;
				}
			}
			buffer[pp] = '0' + mod;
			++ pp;
		}while(x != 0.0F);
	}else if(ip == 0){
		buffer[ip] = '0';
		++ ip;
	}
	length = ip + pp + 1;
	result = new_string(NULL, length);
	result[0] = minus ? '-' : ' ';
	memmove(result + 1 + ip, buffer + ip, pp);
	while(ip > 0){
		result[length - pp - ip] = buffer[ip - 1];
		-- ip;
	}
	return result;
}

DOUBLE VAL(STRING s)
{
	DOUBLE result;
	INTEGER length = LEN(s);
	if(length == 0){
		result = 0.0F;
	}else if(length >= 2 && s[0] == '&' && (s[1] == 'h' || s[1] == 'H')){
		unsigned char *s2 = new_string(s, length);
		s2[0] = '0';
		s2[1] = '0';
		result = strtol((char const *)s2, NULL, 16);
		string_release(s2);
	}else if(length >= 2 && s[0] == '&' && (s[1] == 'o' || s[1] == 'O')){
		unsigned char *s2 = new_string(s, length);
		s2[0] = '0';
		s2[1] = '0';
		result = strtol((char const *)s2, NULL, 8);
		string_release(s2);
	}else{
		result = strtod((char const *)s, NULL);
	}
	return result;
}

/* MATH */

INTEGER power_integer(INTEGER x, INTEGER e)
{
	INTEGER result = 1;
	while(e > 0){
		INTEGER y = result * x;
		if(y < x){ /* crazy... */
			ERROR(overflow);
		}
		result = y;
		-- e;
	}
	return result;
}

SINGLE power_single(SINGLE x, SINGLE e)
{
	return powf(x, e);
}

DOUBLE power_double(DOUBLE x, DOUBLE e)
{
	return pow(x, e);
}

INTEGER SGN(DOUBLE x)
{
	INTEGER result;
	if(x > 0){
		result = 1;
	}else if(x < 0){
		result = -1;
	}else{
		result = 0;
	}
	return result;
}

/* RANDOM */

void RANDOMIZE(INTEGER seed)
{
	srand(seed);
}

INTEGER random_seed(void)
{
	return clock(); /* this is tenuki. */
}

DOUBLE RND(INTEGER action)
{
	static DOUBLE pred = 0.0;

	DOUBLE result;
	if(action < 0){
		srand(action);
	}
	result = pred;
	if(action != 0){
		result = rand() / (RAND_MAX + 1.0);
		pred = result;
	}
	return result;
}

/* DATA */

void read_string(STRING **reading, STRING *item)
{
	string_release(*item);
	*item = **reading;
	if(*item == NULL){
		ERROR(out_of_data);
	}else{
		string_addref(*item);
		++ *reading;
	}
}

void read_double(STRING **reading, DOUBLE *item)
{
	STRING s;
	read_string(reading, &s);
	*item = VAL(s);
	string_release(s);
}

/* I/O */

enum { pos_supported = 8 };
static INTEGER cursor_pos[pos_supported] = {0, };
static INTEGER cursor_line = 0;

static void increment_pos(int f)
{
	if(f < pos_supported){
		++ cursor_pos[f];
	}
}

static void increment_line(int f)
{
	if(f < pos_supported){
		cursor_pos[f] = 0;
	}
	if(f == 1){
		++ cursor_line;
	}
}

static int file_to_posix(INTEGER file, enum open_mode mode)
{
	/* -3:stdin, -2:stdout, -1:stderr, 0:screen, 1...:file */
	int result;
	if(file == 0){
		result = (mode == FOR_INPUT) ? 0 : 1;
	}else if(file < 0){
		result = file + 3;
	}else{
		result = file + 2;
	}
	return result;
}

INTEGER POS(INTEGER file)
{
	int f = file_to_posix(file, FOR_OUTPUT);
	INTEGER result;
	if(f < pos_supported){
		result = cursor_pos[f];
	}else{
		result = 0;
	}
	return result;
}

INTEGER CSRLIN(void)
{
	return cursor_line;
}

void OPEN(STRING filename, enum open_mode mode, INTEGER file)
{
	static int flags[3] = {
		O_TEXT | O_RDONLY,
		O_TEXT | O_WRONLY | O_CREAT | O_TRUNC,
		O_TEXT | O_RDWR | O_CREAT | O_APPEND};

	int n;
	int e;
	int f = file_to_posix(file, FOR_INPUT);
	if(f <= 2){
		ERROR(bad_file_number);
	}
	n = open((char const *)filename, flags[mode], 0777);
	if(n < 0){
		ERROR(disk_io_error);
	}
	/* crazy implementation for OPEN "..." FOR ... AS #n */
	if(n != f){
		close(f);
		e = dup2(n, f);
		close(n);
		if(e < 0){
			ERROR(disk_io_error);
		}
	}
}

void CLOSE(INTEGER file)
{
	int f = file_to_posix(file, FOR_INPUT);
	if(f <= 2){
		ERROR(bad_file_number);
	}
	if(close(f) < 0){
		ERROR(disk_io_error);
	}
}

INTEGER FREEFILE(void)
{
	/* crazy implementation for find dead number */
	int n = dup(0);
	if(n < 0){
		ERROR(disk_io_error);
	}
	close(n);
	return n - 2;
}

LOGICAL EOF(INTEGER file)
{
	struct stat s;
	off_t offset;
	int f = file_to_posix(file, FOR_OUTPUT);
	if(fstat(f, &s) < 0 || (offset = lseek(f, 0, SEEK_CUR)) < 0){
		ERROR(disk_io_error);
	}
	return - (s.st_size == offset);
}

void print_raw(INTEGER file, STRING s)
{
	int f = file_to_posix(file, FOR_OUTPUT);
	int old_mode = setmode(f, O_BINARY);
	if(old_mode < 0){
		ERROR(disk_io_error);
	}
	if(write(f, s, LEN(s)) < 0){
		ERROR(disk_io_error);
	}
	setmode(f, old_mode);
}

void print_character(INTEGER file, CHARACTER c)
{
	int f = file_to_posix(file, FOR_OUTPUT);
	if(c == '\r'){
		c = '\n';
		increment_line(f);
	}else{
		increment_pos(f);
	}
	if(write(f, &c, 1) < 0){
		ERROR(disk_io_error);
	}
}

void print_string(INTEGER file, STRING s)
{
	int f = file_to_posix(file, FOR_OUTPUT);
	INTEGER length = LEN(s);
	INTEGER i;
	for(i = 0; i < length; ++i){
		CHARACTER c = s[i];
		if(c == '\r'){
			c = '\n';
			increment_line(f);
		}else{
			increment_pos(f);
		}
		if(write(f, &c, 1) < 0){
			ERROR(disk_io_error);
		}
	}
}

void print_double(INTEGER file, DOUBLE value)
{
	STRING image = STR(value);
	print_string(file, image);
	string_release(image);
}

void print_line(INTEGER file)
{
	static char const line_feed = '\n';

	int f = file_to_posix(file, FOR_OUTPUT);
	if(write(f, &line_feed, 1) < 0){
		ERROR(disk_io_error);
	}
	increment_line(f);
}

void print_space(INTEGER file, INTEGER count)
{
	static char const space = ' ';

	int f = file_to_posix(file, FOR_OUTPUT);
	INTEGER i;
	for(i = 0; i < count; ++i){
		if(write(f, &space, 1) < 0){
			ERROR(disk_io_error);
		}
		increment_pos(f);
	}
}

void print_tab(INTEGER file, INTEGER count)
{
	static char const space = ' ';

	int f = file_to_posix(file, FOR_OUTPUT);
	INTEGER i;
	for(i = POS(file); i < count; ++i){
		if(write(f, &space, 1) < 0){
			ERROR(disk_io_error);
		}
		increment_pos(f);
	}
}

void print_using_string(INTEGER file, STRING format, INTEGER *index, STRING value)
{
	STRING output = NULL;
	using_string(&output, format, index, value);
	print_string(file, output);
	string_release(output);
}

void print_using_double(INTEGER file, STRING format, INTEGER *index, DOUBLE value)
{
	STRING output = NULL;
	using_double(&output, format, index, value);
	print_string(file, output);
	string_release(output);
}

void print_using_rest(INTEGER file, STRING format, INTEGER *index)
{
	STRING output = NULL;
	using_rest(&output, format, index);
	print_string(file, output);
	string_release(output);
}

void write_string(INTEGER file, STRING s)
{
	INTEGER length;
	INTEGER i;
	print_character(file, '\"');
	length = LEN(s);
	for(i = 0; i < length; ++i){
		print_character(file, s[i]);
		if(s[i] == '\"'){
			print_character(file, '\"');
		}
	}
	print_character(file, '\"');
}

void input_string(INTEGER file, STRING *result)
{
	CHARACTER c;
	INTEGER length = 0;
	int f = file_to_posix(file, FOR_INPUT);
	string_release(*result);
	*result = NULL;
	if(read(f, &c, 1) < 0){
		ERROR(input_past_end);
	}
	while(c <= ' '){
		if(read(f, &c, 1) < 0){
			ERROR(input_past_end);
		}
	}
	if(c == '\"'){
		if(read(f, &c, 1) < 0){
			ERROR(input_past_end);
		}
		for(;;){
			if(c == '\"'){
				if(read(f, &c, 1) < 0){
					break;
				}
				if(c == '\"'){
					++ length;
					set_length(result, length);
					((char *)result)[length - 1] = '\"';
				}else{
					break;
				}
			}else{
				++ length;
				set_length(result, length);
				((char *)result)[length - 1] = c;
			}
			if(read(f, &c, 1) < 0){
				ERROR(input_past_end);
			}
		}
	}else{
		while(c != '\n' && c != ',' && c != ' '){
			++ length;
			set_length(result, length);
			((char *)result)[length - 1] = c;
			if(read(f, &c, 1) < 0){
				break;
			}
		}
	}
	while(c != '\n' && c != ','){
		if(read(f, &c, 1) < 0){
			break;
		}
	}
}

void input_double(INTEGER file, DOUBLE *result)
{
	STRING s = NULL;
	input_string(file, &s);
	*result = VAL(s);
	string_release(s);
}

void line_input(INTEGER file, STRING *result)
{
	CHARACTER c;
	INTEGER length = 0;
	int f = file_to_posix(file, FOR_INPUT);
	string_release(*result);
	*result = NULL;
	for(;;){
		if(read(f, &c, 1) <= 0){
			if(length == 0){
				ERROR(input_past_end);
			}else{
				break;
			}
		}
		if(c == '\n'){
			break;
		}
		++ length;
		set_length(result, length);
		((char *)*result)[length - 1] = c;
	}
}

STRING fnINPUT(INTEGER length, INTEGER file)
{
	int f = file_to_posix(file, FOR_INPUT);
	STRING result = new_string(NULL, length);
	INTEGER r = read(f, (char *)result, length);
	if(r <= 0){
		ERROR(input_past_end);
	}else if(r < length){
		set_length(&result, r);
	}
	return result;
}

/* FILE */

void KILL(STRING filename)
{
	if(filename == NULL || unlink((char const *)filename) < 0){
		ERROR(disk_io_error);
	}
}

/* PROCESS */

INTEGER IARGC(void)
{
	return basic_argc - 1;
}

void GETARG(INTEGER index, STRING *result)
{
	if(index < 0 || index >= basic_argc){
		string_release(*result);
		*result = NULL;
	}else{
		set_string(result, (unsigned char const *)basic_args[index], strlen(basic_args[index]));
	}
}

void GETENV(STRING name, STRING *result)
{
	char const *v = getenv((char const *)name);
	set_string(result, (unsigned char const *)v, strlen(v));
}

INTEGER SHELL(STRING command)
{
	INTEGER result;
	/* "system" accepts NULL */
	if((result = system((char const *)command)) < 0){
		ERROR(disk_io_error);
	}
	return result;
}
