/* BASIC/LL runtime-library */

#ifndef __LIBBASICLL_H__
#define __LIBBASICLL_H__

#define BASIC_CONTROL_STACK

#include <stddef.h>
#include <stdint.h>
#include <setjmp.h>

#if ! defined(__GCC) && ! defined(__GNUC__)
#define __attribute__(x)
#endif

typedef intptr_t INTEGER;
typedef float SINGLE;
typedef double DOUBLE;
typedef unsigned char const *STRING; /* has same representation of Delphi's AnsiString */
typedef unsigned char CHARACTER; /* named by Fortran */
typedef int LOGICAL; /* 0 or -1, named by Fortran */

#define STRING_CONST(name, s) static struct { intptr_t ref_count; intptr_t length; unsigned char data[sizeof(s)]; } const name = {-1, sizeof(s) - 1, s}
#define S(s) (struct { intptr_t ref_count; intptr_t length; unsigned char data[sizeof(s)]; }){.ref_count = -1, .length = sizeof(s) - 1, .data = s}

extern int basic_argc;
extern char **basic_args;
extern int basic_exit_code;
extern intptr_t basic_current_line;
extern jmp_buf basic_on_exit;

/* CONTROL */

#ifdef BASIC_CONTROL_STACK

enum basic_control_kind_t { FOR = 1, WHILE = 2, GOSUB = 3 };

jmp_buf *enter_control(enum basic_control_kind_t kind, void *control);
void leave_control(enum basic_control_kind_t kind, void *control);
void back_control(enum basic_control_kind_t kind, void *control) __attribute__((noreturn));

#endif

void END(void) __attribute__((noreturn));

enum {
	no_error = 0,
	next_without_for = 1,
	syntax_error = 2,
	/* return_without_gosub = 3, */
	out_of_data = 4,
	illegal_function_call = 5,
	overflow = 6,
	out_of_memory = 7,
	/* undefined_line_number = 8, */
	subscript_out_of_range = 9,
	/* redimensioned_array = 10, (duplicate_definition) */
	division_by_zero = 11,
	/* illegal_direct = 12, */
	type_mismatch = 13,
	/* out_of_string_space = 14, */
	string_too_long = 15,
	/* formula_too_complex = 16, */
	/* cannot_continue = 17, */
	/* undefined_user_function = 18, */
	/* device_io_error = 19, */
	illegal_address = 20,
	/* no_resume = 21, */
	/* resume_without_error = 22, */
	/* undefined_label = 23, */
	missing_operand = 24,
	/* line_buffer_overflow = 25, */
	/* ... */
	internal_error = 51,
	bad_file_number = 52,
	file_not_found = 53,
	bad_file_mode = 54,
	input_past_end = 55,
	bad_file_name = 56,
	disk_io_error = 57,
	/* ... */
	user_error = 255,
};

INTEGER ERR(void);
INTEGER ERL(void);
void ERROR(INTEGER number) __attribute__((noreturn));

/* MEMORY */

void *allocate(INTEGER size) __attribute__((__malloc__));
void *reallocate(void *memory, INTEGER size);
void deallocate(void *memory);

/* STRING */

void string_addref(STRING s);
void string_release(STRING s);
unsigned char *new_string(unsigned char const *data, intptr_t length);
void unique_string(STRING *s);
void set_string(STRING *s, unsigned char const *data, intptr_t length);
void set_length(STRING *s, intptr_t length);
INTEGER LEN(STRING s);
INTEGER string_compare(STRING left, STRING right);
STRING string_concatenate(STRING left, STRING right);
CHARACTER const *STRPTR(STRING s);
STRING MID(STRING s, INTEGER index, INTEGER length);
STRING LEFT(STRING s, INTEGER length);
STRING RIGHT(STRING s, INTEGER length);
STRING SPACE(INTEGER count);
STRING fnSTRING(STRING s, INTEGER count);
STRING UCASE(STRING s); /* named by Microsoft BASIC /True BASIC */
STRING LCASE(STRING s); /* named by Microsoft BASIC /True BASIC */

void using_string(STRING *output, STRING format, INTEGER *index, STRING value);
void using_double(STRING *output, STRING format, INTEGER *index, DOUBLE value);
void using_rest(STRING *output, STRING format, INTEGER *index);

#define USING(result, format, values) \
	do{ STRING_CONST(u_format, format); STRING *u_result = result; INTEGER using = 0; \
	values using_rest(u_result, u_format.data, &using); }while(0)
#define US(value) using_string(u_result, u_format.data, &using, value);
#define UD(value) using_double(u_result, u_format.data, &using, value);

#define string_append(s, e) do{ STRING w = string_concatenate(*s, e); string_release(*s); *s = w; }while(0)

/* DATA CONVERTION */

STRING MKI(INTEGER value);
STRING MKS(SINGLE value);
STRING MKD(DOUBLE value);
INTEGER CVI(STRING s);
SINGLE CVS(STRING s);
DOUBLE CVD(STRING s);
STRING STR(DOUBLE x);
DOUBLE VAL(STRING s);

/* MATH */

INTEGER power_integer(INTEGER x, INTEGER e);
SINGLE power_single(SINGLE x, SINGLE e);
DOUBLE power_double(DOUBLE x, DOUBLE e);

INTEGER SGN(DOUBLE x);

/* RANDOM */

void RANDOMIZE(INTEGER seed);
INTEGER random_seed(void);
DOUBLE RND(INTEGER action);

/* DATA */

void read_string(STRING **reading, STRING *item);
void read_double(STRING **reading, DOUBLE *item);

/* I/O */

enum open_mode { FOR_INPUT = 0, FOR_OUTPUT = 1, FOR_APPEND = 2 };

enum { standard_error = -2 };

INTEGER POS(INTEGER file);
INTEGER CSRLIN(void);
void OPEN(STRING filename, enum open_mode mode, INTEGER file);
void CLOSE(INTEGER file);
INTEGER FREEFILE(void); /* named by Microsoft BASIC */
LOGICAL EOF(INTEGER file);

void print_raw(INTEGER file, STRING s);
void print_character(INTEGER file, CHARACTER c);
void print_string(INTEGER file, STRING s);
void print_double(INTEGER file, DOUBLE value);
void print_line(INTEGER file);
void print_space(INTEGER file, INTEGER count);
void print_tab(INTEGER file, INTEGER count);

void print_using_string(INTEGER file, STRING format, INTEGER *index, STRING value);
void print_using_double(INTEGER file, STRING format, INTEGER *index, DOUBLE value);
void print_using_rest(INTEGER file, STRING format, INTEGER *index);

#define PRINT_USING(file, format, values) \
	do{ STRING_CONST(u_format, format); INTEGER u_file = file; INTEGER using = 0; \
	values print_using_rest(u_file, u_format.data, &using); }while(0)
#define PUS(value) print_using_string(u_file, u_format.data, &using, value);
#define PUD(value) print_using_double(u_file, u_format.data, &using, value);

void write_string(INTEGER file, STRING s);

void input_string(INTEGER file, STRING *result);
void input_double(INTEGER file, DOUBLE *result);
void line_input(INTEGER file, STRING *result);

STRING fnINPUT(INTEGER length, INTEGER file);

/* FILE */

void KILL(STRING filename);

/* PROCESS */

INTEGER IARGC(void); /* named by Fortran */
void GETARG(INTEGER index, STRING *result); /* named by Fortran */
void GETENV(STRING name, STRING *result); /* named by Fortran */
INTEGER SHELL(STRING command); /* named by Microsoft BASIC / F-BASIC */

#endif
